// This is the main file for all unit tests using CATCH.
//
// The tests are compiled and executed with "make", further tests can be
// added in the Makefile.
//
// See also: https://github.com/philsquared/Catch/blob/master/docs/tutorial.md

#define CATCH_CONFIG_MAIN
#include "catch/catch.hpp"
