// This file contains documentation about the MimoProcessor class.

/// @file
/// Some documentation about apf::MimoProcessor

/** @page MimoProcessor

For class documentation see apf::MimoProcessor.

@dontinclude dummy_example.cpp

How to use the class MimoProcessor?

You can see the full example there: examples/dummy_example.cpp

First, we include the main header file ...

@skipline mimoprocessor.h

... then some policies (see @ref apf_policies for more information).

@until dummy_thread_policy.h

[class definition]

@until {

[input typedef]

@until Input

[MyIntermediateThing]

@until {

[MyIntermediateThing process]

@until }

@skipline };

[output class]

@until {

[output ctor]

@until }

[Process]

@until }

@skipline };

[ctor]

@until {

[inside ctor]

@until }

[Process]

@until }

[private members]

@until };

And that's it! That's you own class based on apf::MimoProcessor!

**/
